package com.stephengware.java.mss;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.net.Socket;

/**
 * <p>Represents an individual connection to the server. <tt>Connection</tt>s
 * are created by {@link EventQueue}s.</p>
 * 
 * <p>When a new connection is created, a {@link ConnectEvent} will be added to
 * the {@link EventQueue} that created it.</p>
 * 
 * <p>Each connection starts its own thread to listen for input. Any input
 * received will be reported via an {@link InputEvent} in the
 * {@link EventQueue} that created this connection.</p>
 * 
 * <p>Messages can be sent to the connection using the
 * {@link #write(String) write} method.</p>
 * 
 * <p>A connection can be closed via the {@link #close() close} method, or it
 * can be terminated by the client. In either case, a {@link DisconnectEvent}
 * will be added to the {@link EventQueue} that created this connection.</p>
 * 
 * <p>If you want to extend this class, you can override
 * {@link EventQueue#createNewConnection(Socket) EventQueue.createNewConnection}
 * or
 * {@link MultiSocketServer#createNewConnection(Socket) MultiSocketServer.createNewConnection}
 * to return subclasses of this class.</p>
 * 
 * <p><tt>Connection</tt> assumes that messages are segmented by the new line
 * (<tt>\n</tt>) character. In other words, input will not be reported until a
 * new line character is read, and a new line character will be appended to
 * every message that is sent. You can change this behavior by overriding the
 * {@link #readInput() readInput} and {@link #writeOutput(String) writeOutput}
 * methods.</p>
 * 
 * @author Stephen G. Ware
 * @version 1
 */
public class Connection {

	/** The EventQueue which created this connection. */
	private final EventQueue eventQueue;
	
	/** The socket from which this <tt>Connection</tt> reads and to which this
	 * connection writes. */
	protected final Socket socket;
	
	/** Used to read input from the socket. */
	protected final BufferedReader reader;
	
	/** Used to write output to the socket. */
	protected final BufferedWriter writer;
	
	/** A separate thread to listen for input from the socket. */
	final SocketListener socketListener = new SocketListener();
	
	/** Has the connection been closed? */
	private boolean closed = false;
	
	/**
	 * <p>Creates a new <tt>Connection</tt> using the specified reader and
	 * writer.</p>
	 * 
	 * <p>Note that the reader will be wrapped in a {@link BufferedReader} if
	 * it is not already. The same is true for the writer.</p>
	 * 
	 * @param eventQueue the {@link EventQueue} that accepted the socket and is creating this connection
	 * @param socket the socket from which this <tt>Connection</tt> will read and to which this <tt>Connection</tt> will write
	 * @param reader the reader used to read input from the socket
	 * @param writer the writer used to write output to the socket
	 */
	public Connection(EventQueue eventQueue, Socket socket, Reader reader, Writer writer){
		this.eventQueue = eventQueue;
		this.socket = socket;
		if(reader instanceof BufferedReader)
			this.reader = (BufferedReader) reader;
		else
			this.reader = new BufferedReader(reader);
		if(writer instanceof BufferedWriter)
			this.writer = (BufferedWriter) writer;
		else
			this.writer = new BufferedWriter(writer);
		socketListener.start();
	}
	
	/**
	 * Creates a new <tt>Connection</tt> using an {@link InputStreamReader} to
	 * read from the socket and an {@link OutputStreamWriter} to write to the
	 * socket.
	 * 
	 * @param eventQueue
	 * @param socket
	 * @throws IOException
	 */
	public Connection(EventQueue eventQueue, Socket socket) throws IOException {
		this(eventQueue, socket, new InputStreamReader(socket.getInputStream()), new OutputStreamWriter(socket.getOutputStream()));
	}
	
	/**
	 * A separate thread used to listen for input from the socket.
	 * 
	 * @author Stephen G. Ware
	 * @version 1
	 */
	final class SocketListener extends Thread {
		
		/**
		 * Creates a new socket listener.
		 */
		private SocketListener(){
			setName("Connection Listener");
		}
		
		/**
		 * Adds a {@link ConnectEvent} to the EventQueue, then listens for
		 * input from the socket, adding {@link InputEvent}s, and then adds a
		 * {@link DisconnectEvent} to the EventQueue before stopping.
		 */
		public void run(){
			eventQueue.push(new ConnectEvent(Connection.this));
			while(!closed){
				String input = read();
				if(input == null)
					close();
				else
					eventQueue.push(new InputEvent(Connection.this, input));
			}
			eventQueue.push(new DisconnectEvent(Connection.this));
		}
	}
	
	/**
	 * Swallows any IOExceptions thrown during reading.
	 * 
	 * @return the message or null if the socket is closed or if an IOException occurs
	 */
	private final String read(){
		try{
			return readInput();
		}
		catch(IOException ex){
			return null;
		}
	}

	/**
	 * Reads a single message of input from the socket. This method assumes
	 * that a message is ended by a new line character (by using
	 * {@link BufferedReader#readLine()}). You can change this behavior by
	 * overriding this method.
	 * 
	 * @return the message or null if the socket is closed
	 * @throws IOException if an I/O error occurs while reading
	 */
	protected String readInput() throws IOException {
		return reader.readLine();
	}
	
	/**
	 * Writes a message to the socket using
	 * {@link #writeOutput(String) writeOutput}. If an exception is thrown
	 * during writing, it is ignored. Thus, it is always safe to use this
	 * method to send messages (even if the socket is closed), but it
	 * cannot be guaranteed that the message will be received by the client.
	 * 
	 * @param output the message to send
	 */
	public final void write(String output){
		try{
			writeOutput(output);
		}
		catch(IOException ex){
			/* do nothing */
		}
	}
	
	/**
	 * Writes a message to the socket. This method assumes that a message is
	 * ended by a new line character. After the message is sent, a new line
	 * character is also sent and the writer flushed. You can change this
	 * behavior by overriding this method.
	 * 
	 * @param output the message to send
	 * @throws IOException if an I/O error occurs when sending the message
	 */
	protected void writeOutput(String output) throws IOException {
		writer.append(output);
		writer.append('\n');
		writer.flush();
	}
	
	/**
	 * <p>Disconnects this <tt>Connection</tt> from the server. After this
	 * method is called, the <tt>Connection</tt> will no longer listen for
	 * input or be able to send output to the socket, however any
	 * {@link InputEvent}s that were added to the {@link EventQueue} before
	 * this method was called will not be removed from the queue.</p>
	 * 
	 * <p>Whether the <tt>Connection</tt> is closed via this method or closed
	 * by the client, a {@link DisconnectEvent} will always be added to the
	 * {@link EventQueue} that created this connection.</p>
	 */
	public final void close(){
		closed = true;
		try{
			socket.close();
		}
		catch(IOException ex){
			/* do nothing */
		}
	}
}
