package com.stephengware.java.mss;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;

/**
 * <p>An <tt>EventQueue</tt> listens for incoming {@link Connection}s and
 * allows the user to get a sequence of {@link Event}s via the
 * {@link #pop() pop} method.</p>
 * 
 * <p>The sequence of events is guaranteed to have the following properties:</p>
 * <ul>
 * <li>The {@link Event#SERVER_START_EVENT} will always be first (unless the
 * server failed to start).</li>
 * <li>Exactly one {@link ConnectEvent} will occur per connection.</li>
 * <li>Exactly one {@link DisconnectEvent} will occur per connection.
 * There will always be a {@link DisconnectEvent} for every connection,
 * even if the event queue is stopped while the connection is active.
 * <li>Zero to many {@link InputEvent}s will occur per connection.
 * {@link InputEvent}s will always come after the connection's
 * {@link ConnectEvent} and before the connection's {DisconnectEvent}.</li>
 * <li>The {@link Event#SERVER_STOP_EVENT} will always be last (unless the
 * server failed to start).</li>
 * <li>{@link ExceptionEvent}s can occur at any time. If the server fails to
 * start, an {@link ExceptionEvent} will explain why.</li>
 * </ul>
 * 
 * <p>The <tt>EventQueue</tt> does not begin listening for connections until
 * the {@link #start() start} method is called. When the {@link #stop() stop}
 * method is called, all active connections will be closed and any new
 * connections will be refused. The {@link #pop() pop} method will continue to
 * return events until all the requirements above are met.</p>
 * 
 * <p>Calls to {@link #pop() pop} before the queue is started or after the
 * queue has returned the {@link Event#SERVER_STOP_EVENT} will return null.</p>
 * 
 * <p>An <tt>EventQueue</tt> can only be started once.</p>
 * 
 * <p>The <tt>EventQueue</tt> creates new {@link Connection} objects for each
 * new socket that connects. If you want to extend {@link Connection}, override
 * the protected {@link #createNewConnection(Socket) createNewConnection}
 * method.</p>
 * 
 * @author Stephen G. Ware
 * @version 1
 */
public class EventQueue {

	/** The port on which the event queue will listen for connections. */
	public final int port;
	
	/** The queue of events. */
	private final Queue<Event> queue = new LinkedList<>();
	
	/** A separate thread that listens for and creates new connections. */
	private final ConnectionListener connectionListener = new ConnectionListener();
	
	/** A list of currently connected connections. */
	private final HashSet<Connection> connections = new HashSet<>();
	
	/** Has the event queue been started? */
	private boolean started = false;
	
	/** Has the event queue been stopped? */
	private boolean stopped = false;
	
	/** Have all events been popped? */
	private boolean done = false;
	
	/**
	 * Creates a new <tt>EventQueue</tt> which, when started, will listen on 
	 * the given port for new {@link Connection}s and begin filling with
	 * {@link Event}s.
	 *
	 * @param port the port to listen on
	 */
	public EventQueue(int port){
		this.port = port;
	}
	
	/**
	 * A separate thread to listen for and create new connections.
	 * 
	 * @author Stephen G. Ware
	 * @version 1
	 */
	private final class ConnectionListener extends Thread {
		
		/** The {@link ServerSocket} used for listening. */
		private ServerSocket serverSocket = null;
		
		/**
		 * Creates a new listener.
		 */
		private ConnectionListener(){
			setName("Event Queue Connection Listener");
		}
		
		/**
		 * Begins listening and creating connections.
		 */
		public void run(){
			try{
				serverSocket = new ServerSocket(port);
			}
			catch(IOException ex){
				push(new ExceptionEvent(ex));
				return;
			}
			while(!EventQueue.this.stopped){
				Socket socket;
				try{
					socket = serverSocket.accept();
				}
				catch(IOException ex){
					break;
				}
				synchronized(EventQueue.this){
					if(EventQueue.this.stopped){
						try{
							socket.close();
						}
						catch(IOException ex){
							/* do nothing */
						}
					}
					else{
						try{
							connections.add(createNewConnection(socket));
						}
						catch(IOException ex){
							push(new ExceptionEvent(ex));
						}
					}
				}
			}
		}
		
		/**
		 * Stops the listener.
		 */
		private void close(){
			if(serverSocket != null){
				try{
					serverSocket.close();
				}
				catch(IOException ex){
					/* do nothing */
				}
			}
		}
	}
	
	/**
	 * This method creates a new {@link Connection} object for each socket that
	 * connects. If you want to extend {@link Connection}, override this method
	 * to return objects which are a subclass of {@link Connection}.
	 * 
	 * @param socket the newly connected socket
	 * @return a {@link Connection} object for that socket
	 * @throws IOException if an I/O error occurs while creating the {@link Connection}
	 */
	protected Connection createNewConnection(Socket socket) throws IOException {
		return new Connection(this, socket);
	}
	
	/**
	 * When this method is called, the <tt>EventQueue</tt> begins listening on
	 * the specified port for new connections and begins filling with events.
	 */
	public final synchronized void start(){
		if(started)
			throw new IllegalStateException(getClass().getSimpleName() + " can only be started once.");
		started = true;
		this.queue.add(Event.SERVER_START_EVENT);
		connectionListener.start();
	}
	
	/**
	 * Add a new event to the queue and wake some thread which is waiting on
	 * {@link #pop()}.
	 * 
	 * @param event the event to be added
	 */
	synchronized final void push(Event event){
		queue.add(event);
		notify();
	}
	
	/**
	 * Removes the first {@link Event} from the queue and returns it.  If the
	 * queue is empty, this method will block until an event is added.
	 * 
	 * @return the first {@link Event} in the queue
	 */
	public final synchronized Event pop(){
		if(!started || done)
			return null;
		while(queue.isEmpty()){
			try{
				wait();
			}
			catch(InterruptedException ex){
				/* do nothing */
			}
		}
		Event event = queue.poll();
		if(event.type == Event.Type.DISCONNECT){
			connections.remove(((DisconnectEvent) event).connection);
			if(stopped && connections.isEmpty())
				queue.add(Event.SERVER_STOP_EVENT);
		}
		else if(event.type == Event.Type.SERVER_STOP)
			done = true;
		return event;
	}
	
	/**
	 * When this method is called, the <tt>EventQueue</tt> will stop listening
	 * for new connections and close all currently active connections.
	 * The {@link #pop() pop} method will continue to return events after this
	 * method is called until the queue is empty.
	 */
	public synchronized final void stop(){
		if(!started || stopped)
			return;
		stopped = true;
		connectionListener.close();
		if(connections.isEmpty())
			push(Event.SERVER_STOP_EVENT);
		else{
			for(Connection connection : connections)
				connection.close();
		}			
	}
}
