package com.stephengware.java.mss;

import java.io.IOException;
import java.net.Socket;

/**
 * <p>A convenient abstract class which maintains its own {@link EventQueue}
 * and calls the relevant <tt>onX()</tt> method each time a new event is
 * returned from the queue.</p>
 * 
 * <p>The six event methods are as follows:</p>
 * <ul>
 * <li>{@link #onStart() onStart} is called when the server starts.</li>
 * <li>{@link #onConnect(Connection) onConnect} is called each time a new
 * connection is made to the server.</li>
 * <li>{@link #onInput(Connection, String) onInput} is called each time a
 * connection receives input.</li>
 * <li>{@link #onDisconnect(Connection) onDisconnect} is called each time a
 * connections disconnects from the server.</li>
 * <li>{@link #onException(Exception) onException} is called any time an
 * uncaught exception is thrown from one of these six methods or if an
 * {@link IOException} is thrown during the life cycle of a
 * {@link Connection}.</li>
 * <li>{@link #onStop() onStop} is called once all other events have taken
 * place and the server has stopped.</li>
 * </ul>
 * 
 * <p>The order and frequency that these methods are called in reflects the
 * guaranteed properties of the sequence of events returned by
 * {@link EventQueue}.</p>
 * 
 * <p>No two <tt>onX()</tt> methods will ever be called at the same time.
 * The server waits until the currently executing <tt>onX()</tt> returns
 * before calling the next <tt>onX()</tt> method. This means you do not
 * have to worry about synchronized data structures (unless they will be
 * used by other threads).</p>
 * 
 * <p>Note that the server will create a new thread to
 * {@link EventQueue#pop() pop} events from its queue and call the various
 * <tt>onX()</tt> methods. In other words, the <tt>onX()</tt> methods will
 * not be called from the same thread which calls {@link #start() start}.
 * Therefore, it is highly recommended that if you define any new public
 * methods that you make them synchronized. The server's event dispatching
 * thread will always synchronize on this object before calling one of the
 * <tt>onX()</tt> methods.</p>
 * 
 * <p>If you want to extend {@link Connection}, you can override the protected
 * {@link #createNewConnection(Socket) createNewConnection} method to return
 * subclasses of {@link Connection}. This method will be used instead of
 * {@link EventQueue#createNewConnection(Socket) EventQueue.createNewConnection}.
 * </p>
 * 
 * @author Stephen G. Ware
 * @version 1
 */
public abstract class MultiSocketServer {

	/** The port on which this server will listen for connections. */
	public final int port;
	
	/** The server's event queue. */
	private final EventQueue eventQueue;
	
	/**
	 * Creates a new server that will listen on the specified port.
	 * 
	 * @param port the port to listen on
	 */
	public MultiSocketServer(int port){
		this.port = port;
		this.eventQueue = new ServerEventQueue(port);
	}
	
	/**
	 * A special inner subclass of EventQueue which uses the server's
	 * {@link #createNewConnection(Socket)} method.
	 * 
	 * @author Stephen G. Ware
	 */
	private class ServerEventQueue extends EventQueue {

		/**
		 * Creates a new event queue.
		 * 
		 * @param port the port to listen on
		 */
		private ServerEventQueue(int port){
			super(port);
		}
		
		/**
		 * Calls the server's {@link #createNewConnection(Socket)} method.
		 */
		protected Connection createNewConnection(Socket socket) throws IOException {
			return MultiSocketServer.this.createNewConnection(socket);
		}
	}
	
	/**
	 * This method creates a new {@link Connection} object for each socket that
	 * connects. If you want to extend {@link Connection}, override this method
	 * to return objects which are a subclass of {@link Connection}.
	 * 
	 * @param socket the newly connected socket
	 * @return a {@link Connection} object for that socket
	 * @throws IOException if an I/O error occurs while creating the {@link Connection}
	 */
	protected Connection createNewConnection(Socket socket) throws IOException {
		return new Connection(eventQueue, socket);
	}
	
	/**
	 * When this method is called the server will begin listening for
	 * connections on the specified port and calling the appropriate
	 * <tt>onX()</tt> methods. Note that this method does not block;
	 * it returns immediately after being called.
	 */
	public void start(){
		eventQueue.start();
		Event event = eventQueue.pop();
		while(event != null){
			synchronized(this){
				try{
					switch(event.type){
					case SERVER_START:
						onStart();
						break;
					case CONNECT:
						onConnect(((ConnectEvent) event).connection);
						break;
					case INPUT:
						InputEvent inputEvent = (InputEvent) event;
						onInput(inputEvent.connection, inputEvent.input);
						break;
					case DISCONNECT:
						onDisconnect(((DisconnectEvent) event).connection);
						break;
					case EXCEPTION:
						onException(((ExceptionEvent) event).exception);
						break;
					case SERVER_STOP:
						onStop();
						break;
					}
				}
				catch(RuntimeException ex){
					eventQueue.push(new ExceptionEvent(ex));
				}
			}
			event = eventQueue.pop();
		}
	}
	
	/**
	 * Called when the {@link Event#SERVER_START_EVENT} occurs.
	 */
	protected abstract void onStart();
	
	/**
	 * Called when an {@link ConnectEvent} occurs.
	 * 
	 * @param connection the new connection
	 */
	protected abstract void onConnect(Connection connection);
	
	/**
	 * Called when an {@link InputEvent} occurs.
	 * 
	 * @param connection the connection from which the input came
	 * @param input the input from the connection
	 */
	protected abstract void onInput(Connection connection, String input);
	
	/**
	 * Called when an {@link DisconnectEvent} occurs.
	 * 
	 * @param connection the connection which has disconnected
	 */
	protected abstract void onDisconnect(Connection connection);
	
	/**
	 * Called when an {@link ExceptionEvent} occurs.
	 * 
	 * @param exception the exception which occured
	 */
	protected abstract void onException(Exception exception);
	
	/**
	 * Called when the {@link Event#SERVER_STOP_EVENT} occurs.
	 */
	protected abstract void onStop();
	
	/**
	 * When this method is called the server stops accepting new connections
	 * and closes all active connections. Note that the various <tt>onX()</tt>
	 * methods will still be called after this method returns. They will be
	 * called until the server's {@link EventQueue} is empty.
	 */
	public void stop(){
		eventQueue.stop();
	}
}
