/**
 * <p>The Multi-Socket Server or <tt>mss</tt> package provides an easy way to
 * manage asynchronous input from many different sockets by placing all network
 * events into a queue so that they can be handled one at a time on a single
 * thread.</p>
 * 
 * <p>Most servers have at least three kinds of threads:</p>
 * <ol>
 * <li>One main thread which does the actual processing.</li>
 * <li>One thread to listen for and accept new sockets.</li>
 * <li>One thread per socket to listen for input from that socket.</li>
 * </ol>
 * 
 * <p>With so many threads using the same data structures, it can be difficult
 * to avoid race conditions and the other kinds of bugs that crop up in
 * multi-threaded applications, not to mention the difficulty in ensuring that
 * all threads start and end properly. This package provides one solution to
 * that problem.</p>
 * 
 * <p>The main class in this package is
 * {@link com.stephengware.java.mss.EventQueue}, which encapsulates
 * the new socket listener thread (#2 above) and the per-socket listener
 * threads (#3 above). Once the queue is started, it begins to fill with events
 * that represent important network activity. This means you only have to worry
 * about programming the main thread (#1), which can
 * {@link com.stephengware.java.mss.EventQueue#pop() pop} events off of the
 * queue one by one and handle them without worrying about synchronization.</p>
 * 
 * <p>There are six kinds of events that will appear in the queue:</p>
 * <ul>
 * <li>The {@link com.stephengware.java.mss.Event#SERVER_START_EVENT} occurs
 * once when the server starts.</li>
 * <li>A {@link com.stephengware.java.mss.ConnectEvent} occurs each time new
 * socket connects to the server.</li>
 * <li>An {@link com.stephengware.java.mss.InputEvent} occurs each time one of
 * the connected sockets sends a message to the server.</li>
 * <li>A {@link com.stephengware.java.mss.DisconnectEvent} occurs each time 
 * a socket disconnects.</li>
 * <li>The {@link com.stephengware.java.mss.Event#SERVER_START_EVENT} occurs
 * once when the server stops.</li>
 * <li>An {@link com.stephengware.java.mss.ExceptionEvent} occurs any time an
 * uncaught or IOException occurs.</li>
 * </ul>
 * 
 * <p>This package also provides a convenient abstract class, called
 * {@link com.stephengware.java.mss.MultiSocketServer} which can be extended
 * to quickly and easily create a multi-threaded server application. When you
 * extend {@link com.stephengware.java.mss.MultiSocketServer}, you must define
 * six methods, each of which correspond to the six events above:</p>
 * <ul>
 * <li>{@link com.stephengware.java.mss.MultiSocketServer#onStart()} is called
 * when the server starts.</li>
 * <li>{@link com.stephengware.java.mss.MultiSocketServer#onConnect(Connection)}
 * is called each time a new socket connects to the server.</li>
 * <li>{@link com.stephengware.java.mss.MultiSocketServer#onInput(Connection, String)}
 * is called each time a socket receives input.</li>
 * <li>{@link com.stephengware.java.mss.MultiSocketServer#onDisconnect(Connection)}
 * is called each time one of the connected sockets disconnects.</li>
 * <li>{@link com.stephengware.java.mss.MultiSocketServer#onStop()} is called
 * when the server stops.</li>
 * <li>{@link com.stephengware.java.mss.MultiSocketServer#onException(Exception)}
 * is called any time an uncaught exception or IOException occurs.</li>
 * </ul>
 * 
 * <p>These methods will be called one at time as their corresponding events
 * occur, but they will never be called at the same time. In other words, if
 * the server receives a new event while one of the <tt>onX()</tt> methods is
 * executing, it will wait for that method to return before calling the next
 * <tt>onX()</tt> method. This means you don't have to worry about
 * synchronizing the data structures used in your server (unless they will be
 * used by other threads).</p>
 * 
 * <p>Here is a simple example to demonstrate how easy it is to create a chat
 * room server using {@link com.stephengware.java.mss.MultiSocketServer}:</p>
 * 
 * <pre>
 * <code>
 * import com.stephengware.java.mss.*;
 * import java.io.*;
 * import java.util.HashMap;
 * 
 * public class ChatServer extends MultiSocketServer {
 * 
 *      public static final int DEFAULT_PORT = 5555;
 *      protected final HashMap&#60;Connection, String&#62; names = new HashMap&#60;&#62;();
 *      protected final Writer log;
 *      
 *      public static void main(String[] args){
 *           int port = DEFAULT_PORT;
 *           if(args.length &#62; 0){
 *                try{
 *                     port = Integer.parseInt(args[0]);
 *                }
 *                catch(NumberFormatException ex){
 *                     // do nothing
 *                }
 *           }
 *           ChatServer server = new ChatServer(port, new PrintWriter(System.out));
 *           server.start();
 *      }
 *      
 *      public ChatServer(int port, Writer log){
 *           super(port);
 *           this.log = log;
 *      }
 *      
 *      &#64;Override
 *      protected void onStart(){
 *           log("Chat server has started.");
 *      }
 *      
 *      &#64;Override
 *      protected void onConnect(Connection connection){
 *           log("A new client has connected.");
 *           connection.write("Welcome. Please enter your name.");
 *      }
 *      
 *      &#64;Override
 *      protected void onInput(Connection connection, String input){
 *           // If the person already has a name, this input is a chat message.
 *           if(names.containsKey(connection)){
 *                String message = names.get(connection) + ": " + input;
 *                log(message);
 *                writeToAll(message);
 *           }
 *           // If the person does not yet have a name, this input is their name.
 *           else{
 *                log("Identified client \"" + input + "\".");
 *                writeToAll(input + " has joined.");
 *                names.put(connection, input);
 *           }
 *      }
 *      
 *      &#64;Override
 *      protected void onDisconnect(Connection connection){
 *           String name = names.get(connection);
 *           // If they never gave a name, do nothing.
 *           if(name == null)
 *                return;
 *           // If they did give a name, tell the room they have left.
 *           log("Client \"" + name + "\" has disconnected.");
 *           names.remove(connection);
 *           writeToAll(name + " has left.");
 *      }
 *      
 *      &#64;Override
 *      protected void onException(Exception exception){
 *           exception.printStackTrace();
 *      }
 *      
 *      &#64;Override
 *      protected void onStop(){
 *           log("Chat server has stopped.");
 *      }
 *      
 *      protected void log(String message){
 *           try{
 *                log.append(message);
 *                log.append("\n");
 *                log.flush();
 *           }
 *           catch(IOException ex){
 *                ex.printStackTrace();
 *                stop();
 *           }
 *      }
 *      
 *      protected void writeToAll(String message){
 *           for(Connection connection : names.keySet())
 *                connection.write(message);
 *      }
 * }
 * </code>
 * </pre>
 * 
 * @author Stephen G. Ware
 * @version 1
 */
package com.stephengware.java.mss;